/******************************************************************************
  Copyright (c) 1994 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-stdio.h"

#include "bf_register.h"
#include "bytecodes.h"
#include "config.h"
#include "db.h"
#include "functions.h"
#include "list.h"
#include "opcode.h"
#include "program.h"
#include "storage.h"
#include "streams.h"
#include "utils.h"

const char     *mnemonics[256], *ext_mnemonics[256];

struct mapping {
    unsigned	value;
    const char *name;
};

struct mapping mappings[] = {
    {OP_IF,  			"IF"},
    {OP_WHILE,			"WHILE"},
    {OP_EIF,			"ELSEIF"},
    {OP_FORK,			"FORK"},
    {OP_FORK_WITH_ID,		"FORK_NAMED"},
    {OP_FOR_LIST,		"FOR_LIST"},
    {OP_FOR_RANGE,		"FOR_RANGE"},
    {OP_INDEXSET,		"INDEXSET"},
    {OP_PUSH_GET_PROP,		"PUSH_GET_PROP"},
    {OP_GET_PROP,		"GET_PROP"},
    {OP_CALL_VERB,		"CALL_VERB"},
    {OP_PUT_PROP,		"PUT_PROP"},
    {OP_BI_FUNC_CALL,		"CALL_FUNC"},
    {OP_IF_QUES,		"IF_EXPR"},
    {OP_REF,			"INDEX"},
    {OP_RANGE_REF,		"RANGE"},
    {OP_MAKE_SINGLETON_LIST,	"MAKE_SINGLETON_LIST"},
    {OP_CHECK_LIST_FOR_SPLICE,	"CHECK_LIST_FOR_SPLICE"},
    {OP_MULT,			"MULTIPLY"},
    {OP_DIV,			"DIVIDE"},
    {OP_MOD,			"MOD"},
    {OP_ADD,			"ADD"},
    {OP_MINUS,			"SUBTRACT"},
    {OP_EQ,			"EQ"},
    {OP_NE,			"NE"},
    {OP_LT,			"LT"},
    {OP_LE,			"LE"},
    {OP_GT,			"GT"},
    {OP_GE,			"GE"},
    {OP_IN,			"IN"},
    {OP_AND,			"AND"},
    {OP_OR,			"OR"},
    {OP_UNARY_MINUS,		"NEGATE"},
    {OP_NOT,			"NOT"},
    {OP_G_PUT,			"PUT"},
    {OP_G_PUSH,			"PUSH"},
    {OP_IMM,			"PUSH_LITERAL"},
    {OP_MAKE_EMPTY_LIST,	"MAKE_EMPTY_LIST"},
    {OP_LIST_ADD_TAIL,		"LIST_ADD_TAIL"},
    {OP_LIST_APPEND,		"LIST_APPEND"},
    {OP_PUSH_REF,		"PUSH_INDEX"},
    {OP_PUT_TEMP,		"PUT_TEMP"},
    {OP_PUSH_TEMP,		"PUSH_TEMP"},
    {OP_JUMP,			"JUMP"},
    {OP_RETURN,			"RETURN"},
    {OP_RETURN0,		"RETURN 0"},
    {OP_DONE,			"DONE"},
    {OP_POP,			"POP"}};

struct mapping	ext_mappings[] = {{OP_RANGESET, "RANGESET"}};

struct mapping	err_mappings[] = {
    {E_NONE,   	"E_NONE"},
    {E_TYPE,   	"E_TYPE"},
    {E_DIV,    	"E_DIV"},
    {E_PERM,   	"E_PERM"},
    {E_PROPNF, 	"E_PROPNF"},
    {E_VERBNF, 	"E_VERBNF"},
    {E_VARNF,	"E_VARNF"},
    {E_INVIND,	"E_INVIND"},
    {E_RECMOVE,	"E_RECMOVE"},
    {E_MAXREC,	"E_MAXREC"},
    {E_RANGE,	"E_RANGE"},
    {E_ARGS,	"E_ARGS"},
    {E_NACC,	"E_NACC"},
    {E_INVARG,	"E_INVARG"},
    {E_QUOTA,	"E_QUOTA"}};

const char     *error_names[Arraysize(err_mappings, err_mappings[0])];

static void
initialize_tables(void)
{
    static int	tables_initialized = 0;
    unsigned	i;

    if (tables_initialized)
	return;

    for (i = 0; i < 256; i++) {
	mnemonics[i] = "*** Unknown opcode ***";
	ext_mnemonics[i] = "*** Unknown extended opcode ***";
    }

    for (i = 0; i < Arraysize(mappings, mappings[0]); i++)
	mnemonics[mappings[i].value] = mappings[i].name;

    for (i = 0; i < Arraysize(ext_mappings, ext_mappings[0]); i++)
	ext_mnemonics[ext_mappings[i].value] = ext_mappings[i].name;

    for (i = 0; i < Arraysize(err_mappings, err_mappings[0]); i++)
	error_names[err_mappings[i].value] = err_mappings[i].name;

    tables_initialized = 1;
}

static unsigned
add_bytes(Stream *s, Byte *vector, unsigned pc, unsigned length)
{
    unsigned	arg = 0, b;

    while (length--) {
	b = vector[pc++];
	stream_printf(s, " %03d", b);
	arg = (arg << 8) + b;
    }

    return arg;
}

static void
disassemble(Program *prog, void (*print)(const char *, void *), void *data)
{
    Stream	       *s = new_stream(100);
    Stream	       *insn = new_stream(50);
#   define PRINT	(*print)(reset_stream(s), data)
    int			i, l;
    unsigned		pc, max_arg_bytes;
    Bytecodes		bc;
    const char	       *ptr;
    const char	      **names = prog->var_names;
    unsigned		tmp, num_names = prog->num_var_names;
#   define NAMES(i)	(tmp = i,					\
			 tmp < num_names ? names[tmp]			\
					 : "*** Unknown variable ***")
    Var		       *literals = prog->literals;

    initialize_tables();

    stream_printf(s, "First line number: %d", prog->first_lineno); PRINT;

    for (i = -1; i < 0 || i < prog->fork_vectors_size; i++) {
	PRINT;
	if (i == -1) {
	    stream_printf(s, "Main code vector:"); PRINT;
	    stream_printf(s, "================="); PRINT;
	    bc = prog->main_vector;
	} else {
	    stream_printf(s, "Forked code vector %d:", i);
	    l = stream_length(s);
	    PRINT;
	    while (l--)
		stream_add_char(s, '=');
	    PRINT;
	    bc = prog->fork_vectors[i];
	}

	stream_printf(s, "[Bytes for labels = %d, literals = %d, ",
		      bc.numbytes_label, bc.numbytes_literal);
	stream_printf(s, "forks = %d, variables = %d]",
		      bc.numbytes_fork, bc.numbytes_var_name);
	PRINT;
	stream_printf(s, "[Maximum stack size = %d]", bc.max_stack); PRINT;

	max_arg_bytes = bc.numbytes_label;
	if (bc.numbytes_literal > max_arg_bytes)
	    max_arg_bytes = bc.numbytes_literal;
	if (bc.numbytes_fork + bc.numbytes_var_name > max_arg_bytes)
	    max_arg_bytes = bc.numbytes_fork + bc.numbytes_var_name;
	if (bc.numbytes_var_name + bc.numbytes_label > max_arg_bytes)
	    max_arg_bytes = bc.numbytes_var_name + bc.numbytes_label;

	for (pc = 0; pc < bc.size;) {
	    Byte		b = bc.vector[pc];
	    unsigned		num_blanks = max_arg_bytes;
	    unsigned		arg;
#	    define ADD_BYTES(n)	(arg = add_bytes(s, bc.vector, pc, n),	\
				 pc += n,				\
				 num_blanks -= n,			\
				 arg)
	    unsigned		a1, a2;

	    stream_printf(s, "%3d: %03d", pc++, b);
	    stream_add_string(insn, COUNT_TICK(b) ? " * " : "   ");
	    if (IS_OPTIM_NUM_OPCODE(b))
		stream_printf(insn, "NUM %d", OPCODE_TO_OPTIM_NUM(b));
	    else if (IS_PUSH_n(b))
		stream_printf(insn, "PUSH %s", NAMES(PUSH_n_INDEX(b)));
	    else if (IS_PUT_n(b))
		stream_printf(insn, "PUT %s", NAMES(PUT_n_INDEX(b)));
	    else if (b == OP_EXTENDED)
		stream_add_string(insn, ext_mnemonics[ADD_BYTES(1)]);
	    else {
		stream_add_string(insn, mnemonics[b]);
		switch ((Opcode) b) {
		  case OP_IF:		case OP_IF_QUES:	case OP_EIF:
		  case OP_AND:		case OP_OR:		case OP_JUMP:
		  case OP_WHILE:
		    stream_printf(insn, " %d", ADD_BYTES(bc.numbytes_label));
		    break;
		  case OP_FORK:
		    stream_printf(insn, " %d", ADD_BYTES(bc.numbytes_fork));
		    break;
		  case OP_FORK_WITH_ID:
		    a1 = ADD_BYTES(bc.numbytes_fork);
		    a2 = ADD_BYTES(bc.numbytes_var_name);
		    stream_printf(insn, " %d %s", a1, NAMES(a2));
		    break;
		  case OP_FOR_LIST:
		  case OP_FOR_RANGE:
		    a1 = ADD_BYTES(bc.numbytes_var_name);
		    a2 = ADD_BYTES(bc.numbytes_label);
		    stream_printf(insn, " %s %d", NAMES(a1), a2);
		    break;
		  case OP_G_PUSH:
		  case OP_G_PUT:
		    stream_printf(insn, " %s",
				  NAMES(ADD_BYTES(bc.numbytes_var_name)));
		    break;
		  case OP_IMM:
		    {
			Var v;

			v = literals[ADD_BYTES(bc.numbytes_literal)];
			switch (v.type) {
			  case TYPE_OBJ:
			    stream_printf(insn, " #%d", v.v.obj);
			    break;
			  case TYPE_NUM:
			    stream_printf(insn, " %d", v.v.num);
			    break;
			  case TYPE_STR:
			    stream_add_string(insn, " \"");
			    for (ptr = v.v.str; *ptr; ptr++) {
				if (*ptr == '"' || *ptr == '\\')
				    stream_add_char(insn, '\\');
				stream_add_char(insn, *ptr);
			    }
			    stream_add_char(insn, '"');
			    break;
			  case TYPE_ERR:
			    stream_printf(insn, " %s", error_names[v.v.err]);
			    break;
			  default:
			    stream_printf(insn, " <literal type = %d>",
					  v.type);
			    break;
			}
		    }
		    break;
		  case OP_BI_FUNC_CALL:
		    stream_printf(insn, " %s", name_func_by_num(ADD_BYTES(1)));
		  default:
		    break;
		}
	    }

	    while (num_blanks--)
		stream_add_string(s, "    ");
	    stream_add_string(s, reset_stream(insn));
	    PRINT;
	}
    }

    free_stream(s);
    free_stream(insn);

#   undef PRINT
}

static void
print_line(const char *line, void *data)
{
    fprintf(stderr, "%s\n", line);
}

void				/* not static to silence warning */
disassemble_from_debugger(Program *prog)
{
    disassemble(prog, print_line, 0);
}

struct data {
    char      **lines;
    int		used, max;
};

static void
add_line(const char *line, void *data)
{
    struct data	*d = data;

    if (d->used >= d->max) {
	int	new_max = (d->max == 0 ? 20 : d->max * 2);
	char  **new = mymalloc(sizeof(char **) * new_max, M_DISASSEMBLE);
	int	i;

	for (i = 0; i < d->used; i++)
	    new[i] = d->lines[i];
	if (d->lines)
	    myfree(d->lines, M_DISASSEMBLE);
	d->lines = new;
	d->max = new_max;
    }

    d->lines[d->used++] = str_dup(line);
}

static package
bf_disassemble(Var arglist, Byte next, void *vdata, Objid progr)
{
    Objid		oid = arglist.v.list[1].v.obj;
    const char 	       *vname = arglist.v.list[2].v.str;
    Verbdef	       *v;
    struct data		data;
    Var			r;
    int			i;

    if (!valid(oid))
	return make_error_pack(E_INVIND);
    v = find_verbdef(oid, vname);
    free_var(arglist);
    if (!v)
	return make_error_pack(E_VERBNF);
    if (!Verb_perm(progr, v, VP_READ))
	return make_error_pack(E_PERM);

    data.lines = 0;
    data.used = data.max = 0;
    if (v->program)
	disassemble(v->program, add_line, &data);
    r = new_list(data.used);
    for (i = 1; i <= data.used; i++) {
	r.v.list[i].type = TYPE_STR;
	r.v.list[i].v.str = data.lines[i - 1];
    }
    if (data.lines)
	myfree(data.lines, M_DISASSEMBLE);
    return make_var_pack(r);
}

void
register_disassemble(void)
{
    register_function("disassemble", 2, 2, bf_disassemble, TYPE_OBJ, TYPE_STR);
}

/* $Log$
 */
